@@ -10,17 +10,29 @@
 ***********************************************************************/
 
 static char const RCSID[] =
-"$Id: event_tcp.c,v 1.20 2003/01/20 14:15:07 dfs Exp $";
+"$Id: event_tcp.c,v 1.27 2003/09/04 13:43:02 dfs Exp $";
 
 #include "event_tcp.h"
 #include <unistd.h>
 #include <fcntl.h>
 #include <sys/types.h>
 #include <stdlib.h>
+#include <stdio.h>
 #include <errno.h>
 #include <string.h>
 
 static void free_state(EventTcpState *state);
+static void finish_netstring(EventSelector *, int, char *, int, int, void *);
+
+/* Extra chunk of data for reading netstrings */
+typedef struct netstring_extra_t {
+    EventTcpIOFinishedFunc f;
+    int timeout;
+    void *data;
+} netstring_extra;
+
+/* Do not read absurdly-long netstrings */
+#define NETSTRING_MAX 65536
 
 /**********************************************************************
 * %FUNCTION: handle_accept
@@ -45,11 +57,16 @@ handle_accept(EventSelector *es,
     EventTcpAcceptFunc f;
 
     EVENT_DEBUG(("tcp_handle_accept(es=%p, fd=%d, flags=%u, data=%p)\n", es, fd, flags, data));
-    conn = accept(fd, NULL, NULL);
-    if (conn < 0) return;
     f = (EventTcpAcceptFunc) data;
 
-    f(es, conn);
+    /* More than one connection may have come in, so pull them all off! */
+    while (1) {
+	conn = accept(fd, NULL, NULL);
+	if (conn < 0) return;
+	/* Set close-on-exec flag */
+	set_cloexec(conn);
+	f(es, conn);
+    }
 }
 
 /**********************************************************************
@@ -207,6 +224,7 @@ handle_readable(EventSelector *es,
     if (flags & EVENT_FLAG_TIMEOUT) {
 	errno = ETIMEDOUT;
 	if (state->f) {
+	    *(state->cur) = 0;
 	    (state->f)(es, state->socket, state->buf, done,
 		       EVENT_TCP_FLAG_TIMEOUT, state->data);
 	} else {
@@ -232,6 +250,7 @@ handle_readable(EventSelector *es,
 	    flag = (nread) ? EVENT_TCP_FLAG_IOERROR : EVENT_TCP_FLAG_EOF;
 	    /* error or EOF */
 	    if (state->f) {
+		*(state->cur) = 0;
 		(state->f)(es, state->socket, state->buf, done, flag, state->data);
 	    } else {
 		close(state->socket);
@@ -244,6 +263,7 @@ handle_readable(EventSelector *es,
 	if (done >= state->len) {
 	    /* Read enough! */
 	    if (state->f) {
+		*(state->cur) = 0;
 		(state->f)(es, state->socket, state->buf, done,
 			   EVENT_TCP_FLAG_COMPLETE, state->data);
 	    } else {
@@ -294,6 +314,7 @@ handle_readable(EventSelector *es,
 	    flag = EVENT_TCP_FLAG_COMPLETE;
 	}
 	if (state->f) {
+	    *(state->cur) = 0;
 	    (state->f)(es, state->socket, state->buf, done, flag, state->data);
 	} else {
 	    close(state->socket);
@@ -373,6 +394,157 @@ handle_writeable(EventSelector *es,
 
 }
 
+/**********************************************************************
+* %FUNCTION: EventTcp_ReadNetstring
+* %ARGUMENTS:
+*  es -- event selector
+*  socket -- socket to read from
+*  f -- function to call on EOF or when netstring has been read
+*  timeout -- if non-zero, timeout in seconds after which we cancel
+*             operation.
+*  data -- extra data to pass to function f
+* %RETURNS:
+*  A new EventTcpState token or NULL on error
+* %DESCRIPTION:
+*  Sets up a handler to read a netstring.  A netstring looks like this:
+*        len:data  where "len" is the ASCII representation of the decimal
+*                  length of "data".  For example:  7:example
+***********************************************************************/
+EventTcpState *
+EventTcp_ReadNetstring(EventSelector *es,
+		       int socket,
+		       EventTcpIOFinishedFunc f,
+		       int timeout,
+		       void *data)
+{
+    EventTcpState *s;
+    netstring_extra *e = malloc(sizeof(netstring_extra));
+    if (!e) return NULL;
+
+    e->f = f;
+    e->timeout = timeout;
+    e->data = data;
+
+    /* Read the "len:" part and then hand off to data-reading part */
+    s = EventTcp_ReadBuf(es, socket, 16, ':', finish_netstring,
+			 timeout, 0, e);
+    if (!s) {
+	free(e);
+	return NULL;
+    }
+    return s;
+}
+
+
+/**********************************************************************
+* %FUNCTION: finish_netstring (static function)
+* %ARGUMENTS:
+*  es -- event selector
+*  fd -- descriptor we've read from
+*  buf -- buffer containing "len:" part
+*  len -- number of chars read
+*  flag -- result flag
+*  data -- extra data to pass along
+* %RETURNS:
+*  Nothing
+* %DESCRIPTION:
+*  Processes the second part of reading a netstring
+***********************************************************************/
+void
+finish_netstring(EventSelector *es,
+		 int fd,
+		 char *buf,
+		 int len,
+		 int flag,
+		 void *data)
+{
+    netstring_extra *e = (netstring_extra *) data;
+    int readlen;
+    EventTcpState *state;
+
+    /* If something went wrong, pass on to chained function */
+    if (flag != EVENT_TCP_FLAG_COMPLETE) {
+	if (e->f) {
+	    e->f(es, fd, buf, len, flag, e->data);
+	} else {
+	    close(fd);
+	}
+	free(e);
+	return;
+    }
+
+    /* Get length of netstring to read */
+    if ((sscanf(buf, "%d", &readlen) != 1) ||
+	readlen < 0 ||
+	readlen > NETSTRING_MAX) {
+	if (e->f) {
+	    e->f(es, fd, "", 0, EVENT_TCP_FLAG_IOERROR, e->data);
+	} else {
+	    close(fd);
+	}
+	free(e);
+	return;
+    }
+
+    /* Add closing comma */
+    readlen++;
+    state = EventTcp_ReadBuf(es, fd, readlen, -1, e->f, e->timeout,
+			     1, e->data);
+    if (!state) {
+	if (e->f) {
+	    e->f(es, fd, "", 0, EVENT_TCP_FLAG_IOERROR, e->data);
+	} else {
+	    close(fd);
+	}
+    }
+    free(e);
+    return;
+}
+
+/**********************************************************************
+* %FUNCTION: EventTcp_WriteNetstring
+*  es -- event selector
+*  socket -- socket to read from
+*  buf -- buffer to write
+*  len -- number of bytes to write
+*  f -- function to call on EOF or when all bytes have been read
+*  timeout -- timeout after which to cancel operation
+*  data -- extra data to pass to function f.
+* %RETURNS:
+*  A new EventTcpState token or NULL on error
+* %DESCRIPTION:
+*  Sets up a handler to write a buffer to a socket.
+***********************************************************************/
+EventTcpState *
+EventTcp_WriteNetstring(EventSelector *es,
+			int socket,
+			char const *buf,
+			int len,
+			EventTcpIOFinishedFunc f,
+			int timeout,
+			void *data)
+{
+    char *mybuf;
+    int extralen;
+    EventTcpState *state;
+
+    if (len < 0 || len > NETSTRING_MAX) {
+	return NULL;
+    }
+
+    mybuf = malloc(18+len);
+    if (!mybuf) return NULL;
+
+    sprintf(mybuf, "%d:", len);
+    extralen = strlen(mybuf);
+    memcpy(mybuf+extralen, buf, len);
+    mybuf[extralen+len] = ',';
+    state = EventTcp_WriteBuf(es, socket, mybuf,
+			      len+extralen+1, f, timeout, data);
+    free(mybuf);
+    return state;
+}
+
 /**********************************************************************
 * %FUNCTION: EventTcp_ReadBuf
 * %ARGUMENTS:
@@ -425,12 +597,11 @@ EventTcp_ReadBuf(EventSelector *es,
 
     state->socket = socket;
 
-    state->buf = malloc(len);
+    state->buf = malloc(len+1);
     if (!state->buf) {
 	free_state(state);
 	return NULL;
     }
-
     state->cur = state->buf;
     state->len = len;
     state->f = f;
@@ -472,7 +643,7 @@ EventTcp_ReadBuf(EventSelector *es,
 * %RETURNS:
 *  A new EventTcpState token or NULL on error
 * %DESCRIPTION:
-*  Sets up a handler to fill a buffer from a socket.
+*  Sets up a handler to write a buffer to a socket.
 ***********************************************************************/
 EventTcpState *
 EventTcp_WriteBuf(EventSelector *es,
@@ -488,20 +659,28 @@ EventTcp_WriteBuf(EventSelector *es,
     struct timeval t;
 
     EVENT_DEBUG(("EventTcp_WriteBuf(es=%p, socket=%d, len=%d, timeout=%d)\n", es, socket, len, timeout));
-    if (len <= 0) return NULL;
     if (socket < 0) return NULL;
+    if (len <= 0) {
+	if (!f) close(socket);
+	return NULL;
+    }
 
     /* Make sure socket is non-blocking */
     flags = fcntl(socket, F_GETFL, 0);
     if (flags == -1) {
+	if (!f) close(socket);
 	return NULL;
     }
     if (fcntl(socket, F_SETFL, flags | O_NONBLOCK) == -1) {
+	if (!f) close(socket);
 	return NULL;
     }
 
     state = malloc(sizeof(EventTcpState));
-    if (!state) return NULL;
+    if (!state) {
+	if (!f) close(socket);
+	return NULL;
+    }
     memset(state, 0, sizeof(EventTcpState));
 
     state->socket = socket;
@@ -509,6 +688,7 @@ EventTcp_WriteBuf(EventSelector *es,
     state->buf = malloc(len);
     if (!state->buf) {
 	free_state(state);
+	if (!f) close(socket);
 	return NULL;
     }
     memcpy(state->buf, buf, len);
@@ -531,6 +711,7 @@ EventTcp_WriteBuf(EventSelector *es,
 					    (void *) state);
     if (!state->eh) {
 	free_state(state);
+	if (!f) close(socket);
 	return NULL;
     }
 
