@@ -14,7 +14,7 @@
 ***********************************************************************/
 
 static char const RCSID[] =
-"$Id: md-mx-ctrl.c,v 1.17 2004/03/26 19:39:19 dfs Exp $";
+"$Id: md-mx-ctrl.c,v 1.26 2004/08/13 15:09:23 dfs Exp $";
 
 #include "config.h"
 #define SMALLBUF 16384
@@ -40,12 +40,16 @@ static char const RCSID[] =
 #endif
 
 char cmdbuf[SMALLBUF+1];
+int read_stdin = 0;
+FILE *errfp;
 
 static int
 percent_encode(unsigned char *in,
 	       unsigned char *out,
 	       int outlen);
 
+static int process(char const *sock, char const *cmd);
+
 /**********************************************************************
 * %FUNCTION: percent_decode
 * %ARGUMENTS:
@@ -100,7 +104,7 @@ MXCommand(char const *sockname,
 
     fd = socket(AF_LOCAL, SOCK_STREAM, 0);
     if (fd < 0) {
-	fprintf(stderr, "MXCommand: socket: %s\n", strerror(errno));
+	fprintf(errfp, "ERROR MXCommand: socket: %s\n", strerror(errno));
 	return -1;
     }
 
@@ -110,10 +114,10 @@ MXCommand(char const *sockname,
 
     if (connect(fd, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
 	if (errno == EPERM || errno == EACCES) {
-	    fprintf(stderr, "MXCommand: connect: %s\n", strerror(errno));
+	    fprintf(errfp, "ERROR MXCommand: connect: %s\n", strerror(errno));
 	} else {
-	    fprintf(stderr,
-		    "MXCommand: connect: %s: Is multiplexor running?\n",
+	    fprintf(errfp,
+		    "ERROR MXCommand: connect: %s: Is multiplexor running?\n",
 		strerror(errno));
 	}
 	close(fd);
@@ -133,7 +137,7 @@ MXCommand(char const *sockname,
 	nleft -= n;
     }
     if (n < 0) {
-	fprintf(stderr, "MXCommand: write: %s: Is multiplexor running?\n",
+	fprintf(errfp, "ERROR MXCommand: write: %s: Is multiplexor running?\n",
 		strerror(errno));
 	close(fd);
 	return -1;
@@ -152,7 +156,7 @@ MXCommand(char const *sockname,
 	nleft -= n;
     }
     if (n < 0) {
-	fprintf(stderr, "MXCommand: read: %s: Is multiplexor running?\n",
+	fprintf(errfp, "ERROR MXCommand: read: %s: Is multiplexor running?\n",
 		strerror(errno));
 	close(fd);
 	return -1;
@@ -249,7 +253,7 @@ doStatus(char const *sock)
 	return EXIT_FAILURE;
     }
     if (*ans != 'I' && *ans != 'K' && *ans != 'S' && *ans != 'B') {
-	fprintf(stderr, "%s", ans);
+	fprintf(errfp, "ERROR %s", ans);
 	return EXIT_FAILURE;
     }
 
@@ -297,7 +301,7 @@ doBarStatus(char const *sock)
 	return EXIT_FAILURE;
     }
     if (*ans != 'I' && *ans != 'K' && *ans != 'S' && *ans != 'B') {
-	fprintf(stderr, "%s", ans);
+	fprintf(errfp, "ERROR %s", ans);
 	return EXIT_FAILURE;
     }
 
@@ -370,20 +374,52 @@ doBarStatus(char const *sock)
     return EXIT_SUCCESS;
 }
 
-static void
-usage(void)
+static int
+doLoad(char const *sock)
 {
+    char ans[4096];
+    int msgs_0, msgs_1, msgs_5, msgs_10;
+    double avg_0, avg_1, avg_5, avg_10;
+    double mps_0, mps_1, mps_5, mps_10;
+    double ams_0, ams_1, ams_5, ams_10;
+
+    if (MXCommand(sock, "load\n", ans, sizeof(ans)) < 0) {
+	return EXIT_FAILURE;
+    }
+
+    sscanf(ans, "%d %d %d %d %lf %lf %lf %lf %lf %lf %lf %lf",
+	   &msgs_0, &msgs_1, &msgs_5, &msgs_10,
+	   &avg_0, &avg_1, &avg_5, &avg_10,
+	   &ams_0, &ams_1, &ams_5, &ams_10);
+
+    mps_0 = (double) msgs_0 / 10.0;
+    mps_1 = (double) msgs_1 / 60.0;
+    mps_5 = (double) msgs_5 / (5*60.0);
+    mps_10 = (double) msgs_10 / (10*60.0);
+
+    printf("%6s %17s %15s %15s %s\n", "Load:", "Msgs:", "Msgs/Sec:", "Avg ms/scan:", "  Avg Busy Slaves:");
+    printf("%6s %15d %15.2f %15.1f %15.2f\n", "10 Sec", msgs_0, mps_0, ams_0, avg_0);
+    printf("%6s %15d %15.2f %15.1f %15.2f\n", "1 Min", msgs_1, mps_1, ams_1, avg_1);
+    printf("%6s %15d %15.2f %15.1f %15.2f\n", "5 Min", msgs_5, mps_5, ams_5, avg_5);
+    printf("%6s %15d %15.2f %15.1f %15.2f\n", "10 Min", msgs_10, mps_10, ams_10, avg_10);
+    return EXIT_SUCCESS;
+}
+
+static void
+usage(char const *sock)
+{
+    char ans[SMALLBUF];
     fprintf(stderr, "Usage: md-mx-ctrl [options] command\n\n");
     fprintf(stderr, "Options:\n");
     fprintf(stderr, "-h              -- Display help\n");
     fprintf(stderr, "-s path         -- Specify path to multiplexor socket\n");
-    fprintf(stderr, "\nPossible commands:\n");
-    fprintf(stderr, "status          -- Display slave status\n");
-    fprintf(stderr, "rawstatus       -- Display slave status in computer-readable format\n");
-    fprintf(stderr, "barstatus       -- Display slave status as bar graph\n");
-    fprintf(stderr, "reread          -- Re-read filter rules\n");
-    fprintf(stderr, "msgs            -- Number of message processed since startup\n");
-    fprintf(stderr, "histo           -- Print histogram of busy slaves\n");
+    fprintf(stderr, "-i              -- Read commands from stdin\n");
+    if (!sock) {
+	sock = SPOOLDIR "/mimedefang-multiplexor.sock";
+    }
+    if (MXCommand(sock, "help\n", ans, sizeof(ans)) >= 0) {
+	fprintf(stderr, "\nPossible commands:\n%s", ans);
+    }
 }
 
 int
@@ -392,11 +428,16 @@ main(int argc, char *argv[])
     int c;
 
     char *sock = NULL;
+    errfp = stderr;
 
-    while ((c = getopt(argc, argv, "hs:")) != -1) {
+    while ((c = getopt(argc, argv, "hs:i")) != -1) {
 	switch(c) {
+	case 'i':
+	    read_stdin = 1;
+	    errfp = stdout;
+	    break;
 	case 'h':
-	    usage();
+	    usage(sock);
 	    exit(EXIT_SUCCESS);
 	case 's':
 	    if (sock) free(sock);
@@ -407,31 +448,49 @@ main(int argc, char *argv[])
 	    }
 	    break;
 	default:
-	    usage();
+	    usage(sock);
 	    exit(EXIT_FAILURE);
 	}
     }
 
-    if (!argv[optind]) {
-	usage();
+    if (!read_stdin && !argv[optind]) {
+	usage(sock);
 	exit(EXIT_FAILURE);
     }
     if (!sock) {
 	sock = SPOOLDIR "/mimedefang-multiplexor.sock";
     }
 
-    if (!strcmp(argv[optind], "status")) {
-	exit(doStatus(sock));
-    } else if (!strcmp(argv[optind], "barstatus")) {
-	exit(doBarStatus(sock));
-    } else if (!strcmp(argv[optind], "reread")) {
-	exit(doCmd(sock, "reread\n", 0));
-    } else if (!strcmp(argv[optind], "rawstatus")) {
-	exit(doCmd(sock, "status\n", 0));
-    } else if (!strcmp(argv[optind], "msgs")) {
-	exit(doCmd(sock, "msgs\n", 0));
+
+    if (read_stdin) {
+	while(fgets(cmdbuf, SMALLBUF, stdin)) {
+	    process(sock, cmdbuf);
+	    fflush(stdout);
+	}
     } else {
 	buildCmd(argc-optind, argv+optind, cmdbuf, SMALLBUF);
-	exit(doCmd(sock, cmdbuf, 1));
+	exit(process(sock, cmdbuf));
+    }
+    exit(EXIT_SUCCESS);
+}
+
+static int
+process(char const *sock, char const *cmd) {
+    if (!strcmp(cmd, "status\n")) {
+	return doStatus(sock);
+    } else if (!strcmp(cmd, "barstatus\n")) {
+	return doBarStatus(sock);
+    } else if (!strcmp(cmd, "reread\n")) {
+	return doCmd(sock, "reread\n", 0);
+    } else if (!strcmp(cmd, "rawstatus\n")) {
+	return doCmd(sock, "status\n", 0);
+    } else if (!strcmp(cmd, "msgs\n")) {
+	return doCmd(sock, "msgs\n", 0);
+    } else if (!strcmp(cmd, "rawload\n")) {
+	return doCmd(sock, "load\n", 0);
+    } else if (!strcmp(cmd, "load\n")) {
+	return doLoad(sock);
+    } else {
+	return doCmd(sock, cmd, 1);
     }
 }
