@@ -16,7 +16,7 @@
 ***********************************************************************/
 
 static char const RCSID[] =
-"$Id: mimedefang.c,v 1.205 2004/03/09 17:29:20 dfs Exp $";
+"$Id: mimedefang.c,v 1.216 2004/10/27 13:09:43 dfs Exp $";
 
 /* Define this to work around an M$ Outlook bug! */
 /* #define CONVERT_EMBEDDED_CRS_IN_HEADERS 1 */
@@ -103,6 +103,9 @@ int ConserveDescriptors = 0;
 /* Log "eom" run-times */
 int LogTimes = 0;
 
+/* Default Backlog for "listen" */
+static int Backlog = -1;
+
 /* Run as this user */
 static char *user = NULL;
 extern int drop_privs(char const *user, uid_t uid, gid_t gid);
@@ -110,6 +113,9 @@ extern int drop_privs(char const *user, uid_t uid, gid_t gid);
 /* NOQUEUE */
 static char *NOQUEUE = "NOQUEUE";
 
+/* My IP address */
+static char *MyIPAddress;
+
 /* Header name for validating IP addresses */
 static char ValidateHeader[256];
 
@@ -122,6 +128,7 @@ static int NumAdditionalMacros = 0;
 struct privdata {
     char *hostname;		/* Name of connecting host */
     char *hostip;		/* IP address of connecting host */
+    char *myip;                 /* My IP address, from Sendmail macro */
     char *sender;		/* Envelope sender */
     char *firstRecip;		/* Address of first recipient */
     char *dir;			/* Work directory */
@@ -131,7 +138,6 @@ struct privdata {
     int headerFD;		/* File for message headers */
     int cmdFD;			/* File for commands */
     int numContentTypeHeaders;  /* How many Content-Type headers have we seen? */
-    unsigned char headerPresent; /* Already an X-Scanned-By: header found */
     unsigned char validatePresent; /* Saw a relay-address validation header */
     unsigned char suspiciousBody; /* Suspicious characters in message body? */
     unsigned char lastWasCR;	/* Last char of body chunk was CR? */
@@ -143,6 +149,7 @@ static void write_macro_value(SMFICTX *ctx,
 
 static sfsistat cleanup(SMFICTX *ctx);
 static sfsistat mfclose(SMFICTX *ctx);
+static int do_sm_quarantine(SMFICTX *ctx, char const *reason);
 static void remove_working_directory(struct privdata *data);
 
 static char const *MultiplexorSocketName = NULL;
@@ -178,6 +185,8 @@ static int keepFailedDirectories = 0;
 /* Protect mkdir() with mutex? */
 static int protectMkdirWithMutex = 0;
 
+static void set_dsn(SMFICTX *ctx, char *buf2, int code);
+
 #define NO_DELETE_DIR SPOOLDIR "/DO-NOT-DELETE-WORK-DIRS"
 
 #ifdef ENABLE_DEBUGGING
@@ -383,6 +392,7 @@ mfconnect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *sa)
     }
     data->hostname = NULL;
     data->hostip   = NULL;
+    data->myip     = NULL;
     data->sender   = NULL;
     data->firstRecip = NULL;
     data->dir      = NULL;
@@ -392,7 +402,6 @@ mfconnect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *sa)
     data->headerFD = -1;
     data->cmdFD    = -1;
     data->numContentTypeHeaders = 0;
-    data->headerPresent  = 0;
     data->validatePresent = 0;
     data->suspiciousBody = 0;
     data->lastWasCR      = 0;
@@ -464,6 +473,39 @@ mfconnect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *sa)
     data->suspiciousBody = 0;
     data->lastWasCR = 0;
 
+    if (doRelayCheck) {
+	char buf2[SMALLBUF];
+	int n = MXRelayOK(MultiplexorSocketName, buf2, data->hostip,
+			  data->hostname);
+	if (n == 0) {
+	    set_dsn(ctx, buf2, 5);
+	    /* We reject connections from this relay */
+	    cleanup(ctx);
+	    DEBUG_EXIT("mfconnect", __LINE__, "SMFIS_REJECT");
+	    return SMFIS_REJECT;
+	}
+	if (n < 0) {
+	    set_dsn(ctx, buf2, 4);
+	    cleanup(ctx);
+	    DEBUG_EXIT("mfconnect", __LINE__, "SMFIS_TEMPFAIL");
+	    return SMFIS_TEMPFAIL;
+	}
+	if (n == 2) {
+	    set_dsn(ctx, buf2, 2);
+	    cleanup(ctx);
+	    return SMFIS_ACCEPT;
+	}
+	if (n == 3) {
+	    set_dsn(ctx, buf2, 2);
+	    cleanup(ctx);
+	    return SMFIS_DISCARD;
+	}
+	if (n == 1) {
+	    /* Called only in case we need to delay */
+	    set_dsn(ctx, buf2, 2);
+	}
+    }
+
     DEBUG_EXIT("mfconnect", __LINE__, "SMFIS_CONTINUE");
     return SMFIS_CONTINUE;
 }
@@ -517,7 +559,7 @@ envfrom(SMFICTX *ctx, char **from)
     time_t now = time(NULL);
     unsigned long ulnow = (unsigned long) now;
     char *queueid;
-    char *reply, *code, *dsn;
+    char *me;
 
     DEBUG_ENTER("envfrom", __LINE__);
     /* Get the private context */
@@ -625,7 +667,6 @@ envfrom(SMFICTX *ctx, char **from)
     data->fd = -1;
     data->headerFD = -1;
     data->cmdFD = -1;
-    data->headerPresent = 0;
     data->validatePresent = 0;
     data->filterFailed = 0;
     data->numContentTypeHeaders = 0;
@@ -677,86 +718,49 @@ envfrom(SMFICTX *ctx, char **from)
 	write_macro_value(ctx, AdditionalMacros[i]);
     }
 
-    /* Check relaying and sender */
-    if (doRelayCheck) {
-	int n = MXRelayOK(MultiplexorSocketName,
-			  buf2, data->hostip, data->hostname,
-			  data->heloArg, data->dir, data->qid);
-	if (n == 0) {
-	    if (*buf2) {
-		split_on_space3(buf2, &reply, &code, &dsn);
-		percent_decode((unsigned char *) code);
-		percent_decode((unsigned char *) dsn);
-		percent_decode((unsigned char *) reply);
-
-		MD_SMFI_TRY(set_reply, (ctx, "5", code, dsn, reply));
-	    }
-
-	    /* We reject connections from this relay */
-	    cleanup(ctx);
-	    DEBUG_EXIT("envfrom", __LINE__, "SMFIS_REJECT");
-	    return SMFIS_REJECT;
-	}
-	if (n < 0) {
-	    if (*buf2) {
-		split_on_space3(buf2, &reply, &code, &dsn);
-		percent_decode((unsigned char *) code);
-		percent_decode((unsigned char *) dsn);
-		percent_decode((unsigned char *) reply);
-		MD_SMFI_TRY(set_reply, (ctx, "4", code, dsn, reply));
-	    }
-	    cleanup(ctx);
-	    DEBUG_EXIT("envfrom", __LINE__, "SMFIS_TEMPFAIL");
-	    return SMFIS_TEMPFAIL;
-	}
-	if (n == 2) {
-	    cleanup(ctx);
-	    return SMFIS_ACCEPT;
-	}
-	if (n == 3) {
-	    cleanup(ctx);
-	    return SMFIS_DISCARD;
-	}
+    /* Get my IP address */
+    me = smfi_getsymval(ctx, "{if_addr}");
+    if (me && *me && !strcmp(me, "127.0.0.1")) {
+	data->myip = strdup_with_log(me);
+    } else {
+	/* Sigh... use our computed address */
+	data->myip = MyIPAddress;
     }
+
+
     if (doSenderCheck) {
 	int n = MXSenderOK(MultiplexorSocketName, buf2,
 			   from[0], data->hostip, data->hostname,
 			   data->heloArg, data->dir, data->qid);
 	if (n == 0) {
-	    if (*buf2) {
-		split_on_space3(buf2, &reply, &code, &dsn);
-		percent_decode((unsigned char *) code);
-		percent_decode((unsigned char *) dsn);
-		percent_decode((unsigned char *) reply);
+	    set_dsn(ctx, buf2, 5);
 
-		MD_SMFI_TRY(set_reply, (ctx, "5", code, dsn, reply));
-	    }
 	    /* We reject connections from this sender */
 	    cleanup(ctx);
 	    DEBUG_EXIT("envfrom", __LINE__, "SMFIS_REJECT");
 	    return SMFIS_REJECT;
 	}
 	if (n < 0) {
-	    if (*buf2) {
-		split_on_space3(buf2, &reply, &code, &dsn);
-		percent_decode((unsigned char *) code);
-		percent_decode((unsigned char *) dsn);
-		percent_decode((unsigned char *) reply);
+	    set_dsn(ctx, buf2, 4);
 
-		MD_SMFI_TRY(set_reply, (ctx, "4", code, dsn, reply));
-	    }
 	    cleanup(ctx);
 	    DEBUG_EXIT("envfrom", __LINE__, "SMFIS_TEMPFAIL");
 	    return SMFIS_TEMPFAIL;
 	}
 	if (n == 2) {
+	    set_dsn(ctx, buf2, 2);
 	    cleanup(ctx);
 	    return SMFIS_ACCEPT;
 	}
 	if (n == 3) {
+	    set_dsn(ctx, buf2, 2);
 	    cleanup(ctx);
 	    return SMFIS_DISCARD;
 	}
+	if (n == 1) {
+	    /* Called only in case we need to delay */
+	    set_dsn(ctx, buf2, 2);
+	}
     }
 
     if (queueid) {
@@ -794,7 +798,6 @@ rcptto(SMFICTX *ctx, char **to)
     char ans[SMALLBUF];
     sfsistat retcode = SMFIS_CONTINUE;
     char const *rcpt_mailer, *rcpt_host, *rcpt_addr;
-    char *reply, *code, *dsn;
 
     DEBUG_ENTER("rcptto", __LINE__);
     if (!data) {
@@ -830,34 +833,32 @@ rcptto(SMFICTX *ctx, char **to)
 			  rcpt_mailer, rcpt_host, rcpt_addr);
 	if (n == 0) {
 	    /* We reject to this recipient */
-	    if (*ans) {
-		split_on_space3(ans, &reply, &code, &dsn);
-		percent_decode((unsigned char *) code);
-		percent_decode((unsigned char *) dsn);
-		percent_decode((unsigned char *) reply);
-		MD_SMFI_TRY(set_reply, (ctx, "5", code, dsn, reply));
-	    }
+	    set_dsn(ctx, ans, 5);
+
 	    DEBUG_EXIT("rcptto", __LINE__, "SMFIS_REJECT");
 	    return SMFIS_REJECT;
 	}
 	if (n < 0) {
-	    if (*ans) {
-		split_on_space3(ans, &reply, &code, &dsn);
-		percent_decode((unsigned char *) code);
-		percent_decode((unsigned char *) dsn);
-		percent_decode((unsigned char *) reply);
-		MD_SMFI_TRY(set_reply, (ctx, "4", code, dsn, reply));
-	    }
+	    set_dsn(ctx, ans, 4);
+
 	    DEBUG_EXIT("rcptto", __LINE__, "SMFIS_TEMPFAIL");
 	    return SMFIS_TEMPFAIL;
 	}
 	if (n == 2) {
+	    set_dsn(ctx, ans, 2);
+
 	    retcode = SMFIS_ACCEPT;
 	}
 	if (n == 3) {
+	    set_dsn(ctx, ans, 2);
+
 	    cleanup(ctx);
 	    return SMFIS_DISCARD;
 	}
+	if (n == 1) {
+	    /* Called only in case we need to delay */
+	    set_dsn(ctx, ans, 2);
+	}
     }
 
     /* Write recipient line, only for recipients we accept! */
@@ -963,14 +964,6 @@ header(SMFICTX *ctx, char *headerf, char *headerv)
 	write_mx_command(data->cmdFD, 'X', (unsigned char *) headerv);
     }
 
-    /* Check for existing identical X-Scanned-By: header */
-    if (scan_body && *scan_body) {
-	if (!strcasecmp(headerf, "X-Scanned-By") &&
-	    !strcasecmp(headerv, scan_body)) {
-	    data->headerPresent = 1;
-	}
-    }
-
     /* Check for validating IP header.  If found, write a J line
        to the file to reset the SMTP host address */
     if (ValidateHeader[0] && !strcmp(headerf, ValidateHeader)) {
@@ -1146,6 +1139,7 @@ eom(SMFICTX *ctx)
     char result[SMALLBUF];
     char *rbuf, *rptr, *eptr;
 
+    int seen_F = 0;
     int res_fd;
     int n;
     struct privdata *data = DATA;
@@ -1359,10 +1353,6 @@ eom(SMFICTX *ctx)
 		MD_SMFI_TRY(smfi_chgheader, (ctx, "Content-Type", 1, rptr+1));
 	    }
 	    MD_SMFI_TRY(smfi_chgheader, (ctx, "MIME-Version", 1, "1.0"));
-	    /* We've just changed the message to multi-part.  Better delete
-	       other MIME headers which may be present */
-	    MD_SMFI_TRY(smfi_chgheader, (ctx, "Content-Disposition", 1, NULL));
-	    MD_SMFI_TRY(smfi_chgheader, (ctx, "Content-Transfer-Encoding", 1, NULL));
 	    break;
 
 	case 'H':
@@ -1408,6 +1398,12 @@ eom(SMFICTX *ctx)
 	    MD_SMFI_TRY(smfi_addrcpt, (ctx, rptr+1));
 	    break;
 
+	case 'Q':
+	    /* Quarantine a message using Sendmail's facility */
+	    percent_decode((unsigned char *) rptr+1);
+	    MD_SMFI_TRY(do_sm_quarantine, (ctx, rptr+1));
+	    break;
+
 	case 'S':
 	    /* Delete a recipient */
 	    percent_decode((unsigned char *) rptr+1);
@@ -1415,6 +1411,7 @@ eom(SMFICTX *ctx)
 	    break;
 
 	case 'F':
+	    seen_F = 1;
 	    /* We're done */
 	    break;
 
@@ -1425,8 +1422,18 @@ eom(SMFICTX *ctx)
 	if (*rptr == 'F') break;
     }
 
-    if (!data->headerPresent) {
-	if (scan_body && *scan_body) {
+    if (!seen_F) {
+	syslog(LOG_ERR, "%s: RESULTS file did not finish with 'F' line: Tempfailing",
+	       data->qid);
+	r = SMFIS_TEMPFAIL;
+	goto bail_out;
+    }
+    if (scan_body && *scan_body) {
+	if (data->myip) {
+	    snprintf(buffer, SMALLBUF, "%s on %s", scan_body, data->myip);
+	    buffer[SMALLBUF-1] = 0;
+	    MD_SMFI_TRY(smfi_addheader, (ctx, "X-Scanned-By", buffer));
+	} else {
 	    MD_SMFI_TRY(smfi_addheader, (ctx, "X-Scanned-By", scan_body));
 	}
     }
@@ -1494,6 +1501,7 @@ mfclose(SMFICTX *ctx)
 	if (data->dir)           free(data->dir);
 	if (data->hostname)      free(data->hostname);
 	if (data->hostip)        free(data->hostip);
+	if (data->myip && data->myip != MyIPAddress) free(data->myip);
 	if (data->sender)        free(data->sender);
 	if (data->firstRecip)    free(data->firstRecip);
 	if (data->heloArg)       free(data->heloArg);
@@ -1588,13 +1596,19 @@ static struct smfiDesc filterDescriptor =
     "MIMEDefang-" VERSION,      /* Filter name */
     SMFI_VERSION,		/* Version code */
 
-#if SMFI_VERSION == 1
-    /* We can: add a header and may alter body and add/delete recipients*/
-    SMFIF_MODHDRS|SMFIF_MODBODY|SMFIF_ADDRCPT|SMFIF_DELRCPT,
+#if SMFI_VERSION == 2
+#ifdef SMFIF_QUARANTINE
+    /* We can: quarantine, add a header and may alter body
+       and add/delete recipients*/
+    SMFIF_ADDHDRS|SMFIF_CHGBODY|SMFIF_ADDRCPT|SMFIF_DELRCPT|SMFIF_CHGHDRS|SMFIF_QUARANTINE,
 #else
     /* We can: Add headers, alter body, add/delete recipients, alter headers */
     SMFIF_ADDHDRS|SMFIF_CHGBODY|SMFIF_ADDRCPT|SMFIF_DELRCPT|SMFIF_CHGHDRS,
 #endif
+#elif SMFI_VERSION == 1
+    /* We can: add a header and may alter body and add/delete recipients*/
+    SMFIF_MODHDRS|SMFIF_MODBODY|SMFIF_ADDRCPT|SMFIF_DELRCPT,
+#endif
 
     mfconnect,			/* connection */
     helo,			/* HELO */
@@ -1624,6 +1638,7 @@ usage(void)
     fprintf(stderr, "Usage: mimedefang [options]\n");
     fprintf(stderr, "Options:\n");
     fprintf(stderr, "  -h                -- Print usage info and exit\n");
+    fprintf(stderr, "  -v                -- Print version and exit\n");
     fprintf(stderr, "  -U user           -- Run as user instead of root\n");
     fprintf(stderr, "  -p /path          -- Path to UNIX-domain socket for sendmail communication\n");
     fprintf(stderr, "  -d                -- Enable debugging (do not remove spool files)\n");
@@ -1634,6 +1649,7 @@ usage(void)
     fprintf(stderr, "  -t                -- Do recipient checks before processing body (requires -m)\n");
     fprintf(stderr, "  -P file           -- Write process-ID of daemon to specified file\n");
     fprintf(stderr, "  -T                -- Log filter times to syslog\n");
+    fprintf(stderr, "  -b n              -- Set listen() backlog to n\n");
     fprintf(stderr, "  -C                -- Try very hard to conserve file descriptors\n");
     fprintf(stderr, "  -x string         -- Add string as X-Scanned-By header\n");
     fprintf(stderr, "  -X                -- Do not add X-Scanned-By header\n");
@@ -1664,6 +1680,7 @@ main(int argc, char **argv)
     FILE *fp;
     int facility = LOG_MAIL;
     int nodaemon = 0;
+    char buf[SMALLBUF];
 
     /* Paranoia time */
     umask(027);
@@ -1681,13 +1698,45 @@ main(int argc, char **argv)
 	exit(EXIT_FAILURE);
     }
 
+    MyIPAddress = NULL;
+
+    /* Determine my IP address */
+    if (gethostname(buf, sizeof(buf)) >= 0) {
+	struct hostent *he = gethostbyname(buf);
+	struct in_addr in;
+	if (he && he->h_addr) {
+	    memcpy(&in.s_addr, he->h_addr, sizeof(in.s_addr));
+#ifdef HAVE_INET_NTOP
+	    if (inet_ntop(AF_INET, &in.s_addr, buf, sizeof(buf))) {
+		if (*buf) MyIPAddress = strdup_with_log(buf);
+	    }
+#else
+	    {
+		char *s = inet_ntoa(in.sin_addr);
+		if (s && *s) MyIPAddress = strdup_with_log(s);
+	    }
+#endif
+	} else {
+	    syslog(LOG_WARNING, "Could not determine my own IP address!  Ensure that %s has an entry in /etc/hosts or the DNS", buf);
+	    fprintf(stderr, "Could not determine my own IP address!  Ensure that %s has an entry in /etc/hosts or the DNS\n", buf);
+	}
+    }
+
     /* Process command line options */
-    while ((c = getopt(argc, argv, "a:Chp:dm:srtkP:U:Tx:MXS:D")) != -1) {
+    while ((c = getopt(argc, argv, "a:Chp:dm:srtkP:U:Tx:MXS:Dvb:")) != -1) {
 	switch (c) {
+	case 'b':
+	    sscanf(optarg, "%d", &Backlog);
+	    if (Backlog < 5) Backlog = 5;
+	    break;
 	case 'C':
 	    ConserveDescriptors = 1;
 	    break;
 
+	case 'v':
+	    printf("mimedefang version %s\n", VERSION);
+	    exit(0);
+
 	case 'D':
 	    nodaemon = 1;
 	    break;
@@ -1819,6 +1868,10 @@ main(int argc, char **argv)
 	scan_body = SCAN_BODY;
     }
 
+    if (Backlog > 0) {
+	smfi_setbacklog(Backlog);
+    }
+
     if (!MultiplexorSocketName) {
 	fprintf(stderr, "%s: You must use the '-m' option.\n", argv[0]);
 	exit(EXIT_FAILURE);
@@ -2012,3 +2065,60 @@ remove_working_directory(struct privdata *data)
     }
 }
 
+/**********************************************************************
+* %FUNCTION: set_dsn
+* %ARGUMENTS:
+*  data -- our private data area
+*  ctx -- Milter context
+*  buf2 -- return from a relay/sender/filter check.  Consists of
+*    space-separated "reply code dsn sleep_amount" list.
+*  num -- 0, 4 or 5 -- if 4 or 5, we use the code and dsn in set_reply.
+* %RETURNS:
+*  Nothing
+* %DESCRIPTION:
+*  Sets SMTP reply and possibly delays
+***********************************************************************/
+static void
+set_dsn(SMFICTX *ctx, char *buf2, int num) {
+    char *reply, *code, *dsn, *sleepstr;
+
+    if (*buf2) {
+	split_on_space4(buf2, &reply, &code, &dsn, &sleepstr);
+	percent_decode((unsigned char *) code);
+	percent_decode((unsigned char *) dsn);
+	percent_decode((unsigned char *) reply);
+	percent_decode((unsigned char *) sleepstr);
+	do_delay(sleepstr);
+	if (num == 4 || num == 5) {
+	    struct privdata *data = DATA;
+	    if (num == 5) {
+		MD_SMFI_TRY(set_reply, (ctx, "5", code, dsn, reply));
+	    } else {
+		MD_SMFI_TRY(set_reply, (ctx, "4", code, dsn, reply));
+	    }
+	}
+    }
+}
+
+/**********************************************************************
+* %FUNCTION: do_sm_quarantine
+* %ARGUMENTS:
+*  ctx -- Milter context
+*  reason -- reason for quarantine
+* %RETURNS:
+*  Whatever smfi_quarantine returns
+* %DESCRIPTION:
+*  Quarantines a message using Sendmail's quarantine facility, if supported.
+***********************************************************************/
+static int
+do_sm_quarantine(SMFICTX *ctx,
+		 char const *reason)
+{
+#ifdef SMFIF_QUARANTINE
+    return smfi_quarantine(ctx, reason);
+#else
+    syslog(LOG_WARNING, "smfi_quarantine not supported: Requires Sendmail 8.13.0 or later");
+    return MI_FAILURE;
+#endif
+
+}
