@@ -13,7 +13,7 @@
 ***********************************************************************/
 
 static char const RCSID[] =
-"$Id: utils.c,v 1.36 2004/03/26 19:33:33 dfs Exp $";
+"$Id: utils.c,v 1.42 2004/09/22 14:06:03 dfs Exp $";
 
 #define _BSD_SOURCE 1
 
@@ -119,6 +119,50 @@ split_on_space3(char *buf,
 	}
     }
 }
+
+/**********************************************************************
+* %FUNCTION: split_on_space4
+* %ARGUMENTS:
+*  buf -- input buffer
+*  first -- set to first word
+*  second -- set to second word or NULL
+*  third -- set to third word or NULL
+*  rest -- set to everything following a space, or NULL if no space
+* %RETURNS:
+*  Nothing
+* %DESCRIPTION:
+*  Splits a line on whitespace.
+***********************************************************************/
+void
+split_on_space4(char *buf,
+		char **first,
+		char **second,
+		char **third,
+		char **rest)
+{
+    *first = buf;
+    *second = NULL;
+    *third = NULL;
+    *rest = NULL;
+    while(*buf && !isspace(*buf)) buf++;
+    if (*buf && isspace(*buf)) {
+	*buf = 0;
+	*second = buf+1;
+	buf++;
+	while(*buf && !isspace(*buf)) buf++;
+	if (*buf && isspace(*buf)) {
+	    *buf = 0;
+	    *third = buf+1;
+	    buf++;
+	    while(*buf && !isspace(*buf)) buf++;
+	    if (*buf && isspace(*buf)) {
+		*buf = 0;
+		*rest = buf+1;
+	    }
+	}
+    }
+}
+
 /**********************************************************************
 * %FUNCTION: malloc_with_log
 * %ARGUMENTS:
@@ -390,6 +434,10 @@ percent_decode(unsigned char *buf)
     unsigned char *out = buf;
     unsigned int val;
 
+    if (!buf) {
+	return;
+    }
+
     while(*in) {
 	if (*in == '%' && isxdigit(*(in+1)) && isxdigit(*(in+2))) {
 	    sscanf(in+1, "%2x", &val);
@@ -503,47 +551,21 @@ percent_encode_command(char *out, int outlen, ...)
     return 0;
 }
 
+
 /**********************************************************************
-* %FUNCTION: MXRelayOK
+* %FUNCTION: munch_mx_return
 * %ARGUMENTS:
-*  sockname -- multiplexor socket name
+*  ans -- answer from multiplexor
 *  msg -- buffer for holding error message, at least SMALLBUF chars
-*  ip -- relay IP address
-*  name -- relay name
-*  helo -- argument to "HELO/EHLO" (may be NULL)
-*  dir -- MIMEDefang working directory
-*  qid -- Sendmail queue identifier
 * %RETURNS:
 *  1 if it's OK to accept connections from this host; 0 if not, -1 if error.
 *  If connection is rejected, error message *may* be set.
 ***********************************************************************/
-int
-MXRelayOK(char const *sockname,
-	  char *msg,
-	  char const *ip,
-	  char const *name,
-	  char const *helo,
-	  char const *dir,
-	  char const *qid)
+static int
+munch_mx_return(char *ans, char *msg)
 {
-    char cmd[SMALLBUF];
-    char ans[SMALLBUF];
-    int len;
+    size_t len;
 
-    *msg = 0;
-
-    if (!ip || !*ip) {
-	ip = "UNKNOWN";
-    }
-    if (!name || !*name) {
-	name = ip;
-    }
-    if (!helo) {
-	helo = "UNKNOWN";
-    }
-    percent_encode_command(cmd, sizeof(cmd), "relayok", ip, name, helo,
-			   dir, qid, NULL);
-    if (MXCommand(sockname, cmd, ans, SMALLBUF-1) < 0) return -1;
     if (!strcmp(ans, "ok -1\n")) return -1;
     if (!strcmp(ans, "ok 1\n")) return 1;
     if (!strcmp(ans, "ok 2\n")) return 2;
@@ -552,7 +574,7 @@ MXRelayOK(char const *sockname,
 
     chomp(ans);
 
-    /* If rejection message is supplied, set it */
+    /* If rejection message is supplied, set failure code and return 0 */
     len = strlen(ans);
     if (len >= 6 && !strncmp(ans, "ok 0 ", 5)) {
 	strcpy(msg, ans+5);
@@ -564,10 +586,59 @@ MXRelayOK(char const *sockname,
 	return -1;
     }
 
+    if (len >= 6 && !strncmp(ans, "ok 1 ", 5)) {
+	strcpy(msg, ans+5);
+	return 1;
+    }
+
+    if (len >= 6 && !strncmp(ans, "ok 2 ", 5)) {
+	strcpy(msg, ans+5);
+	return 2;
+    }
+
+    if (len >= 6 && !strncmp(ans, "ok 3 ", 5)) {
+	strcpy(msg, ans+5);
+	return 3;
+    }
+
     if (len > 0 && ans[len-1] == '\n') ans[len-1] = 0;
     syslog(LOG_ERR, "Error from multiplexor: %s", ans);
     return -1;
+}
 
+/**********************************************************************
+* %FUNCTION: MXRelayOK
+* %ARGUMENTS:
+*  sockname -- multiplexor socket name
+*  msg -- buffer for holding error message, at least SMALLBUF chars
+*  ip -- relay IP address
+*  name -- relay name
+* %RETURNS:
+*  1 if it's OK to accept connections from this host; 0 if not, -1 if error.
+*  If connection is rejected, error message *may* be set.
+***********************************************************************/
+int
+MXRelayOK(char const *sockname,
+	  char *msg,
+	  char const *ip,
+	  char const *name)
+{
+    char cmd[SMALLBUF];
+    char ans[SMALLBUF];
+
+    *msg = 0;
+
+    if (!ip || !*ip) {
+	ip = "UNKNOWN";
+    }
+    if (!name || !*name) {
+	name = ip;
+    }
+    if (percent_encode_command(cmd, sizeof(cmd), "relayok", ip, name, NULL) < 0) {
+	return -1;
+    }
+    if (MXCommand(sockname, cmd, ans, SMALLBUF-1) < 0) return -1;
+    return munch_mx_return(ans, msg);
 }
 
 /**********************************************************************
@@ -598,7 +669,6 @@ MXSenderOK(char const *sockname,
 {
     char cmd[SMALLBUF];
     char ans[SMALLBUF];
-    int len;
 
     *msg = 0;
 
@@ -615,32 +685,12 @@ MXSenderOK(char const *sockname,
     if (!helo) {
 	helo = "UNKNOWN";
     }
-    percent_encode_command(cmd, sizeof(cmd), "senderok", sender, ip, name,
-			   helo, dir, qid, NULL);
-    if (MXCommand(sockname, cmd, ans, SMALLBUF-1) < 0) return -1;
-    if (!strcmp(ans, "ok -1\n")) return -1;
-    if (!strcmp(ans, "ok 1\n")) return 1;
-    if (!strcmp(ans, "ok 2\n")) return 2;
-    if (!strcmp(ans, "ok 3\n")) return 3;
-    if (!strcmp(ans, "ok 0\n")) return 0;
-
-    chomp(ans);
-
-    /* If rejection message is supplied, set failure code and return 0 */
-    len = strlen(ans);
-    if (len >= 6 && !strncmp(ans, "ok 0 ", 5)) {
-	strcpy(msg, ans+5);
-	return 0;
-    }
-
-    if (len >= 7 && !strncmp(ans, "ok -1 ", 6)) {
-	strcpy(msg, ans+6);
+    if (percent_encode_command(cmd, sizeof(cmd), "senderok", sender, ip, name,
+			       helo, dir, qid, NULL) < 0) {
 	return -1;
     }
-
-    if (len > 0 && ans[len-1] == '\n') ans[len-1] = 0;
-    syslog(LOG_ERR, "Error from multiplexor: %s", ans);
-    return -1;
+    if (MXCommand(sockname, cmd, ans, SMALLBUF-1) < 0) return -1;
+    return munch_mx_return(ans, msg);
 }
 
 /**********************************************************************
@@ -681,7 +731,6 @@ MXRecipientOK(char const *sockname,
 {
     char cmd[SMALLBUF];
     char ans[SMALLBUF];
-    int len;
 
     *msg = 0;
 
@@ -707,34 +756,14 @@ MXRecipientOK(char const *sockname,
 	helo = "UNKNOWN";
     }
 
-    percent_encode_command(cmd, sizeof(cmd),
-			   "recipok", recipient, sender, ip, name, firstRecip,
-			   helo, dir, qid, rcpt_mailer, rcpt_host, rcpt_addr,
-			   NULL);
-    if (MXCommand(sockname, cmd, ans, SMALLBUF-1) < 0) return -1;
-    if (!strcmp(ans, "ok -1\n")) return -1;
-    if (!strcmp(ans, "ok 1\n")) return 1;
-    if (!strcmp(ans, "ok 2\n")) return 2;
-    if (!strcmp(ans, "ok 3\n")) return 3;
-    if (!strcmp(ans, "ok 0\n")) return 0;
-
-    chomp(ans);
-
-    /* If rejection message is supplied, set failure code and return 0 */
-    len = strlen(ans);
-    if (len >= 6 && !strncmp(ans, "ok 0 ", 5)) {
-	strcpy(msg, ans+5);
-	return 0;
-    }
-
-    if (len >= 7 && !strncmp(ans, "ok -1 ", 6)) {
-	strcpy(msg, ans+6);
+    if (percent_encode_command(cmd, sizeof(cmd),
+			       "recipok", recipient, sender, ip, name, firstRecip,
+			       helo, dir, qid, rcpt_mailer, rcpt_host, rcpt_addr,
+			       NULL) < 0) {
 	return -1;
     }
-
-    if (len > 0 && ans[len-1] == '\n') ans[len-1] = 0;
-    syslog(LOG_ERR, "Error from multiplexor: %s", ans);
-    return -1;
+    if (MXCommand(sockname, cmd, ans, SMALLBUF-1) < 0) return -1;
+    return munch_mx_return(ans, msg);
 }
 
 /**********************************************************************
@@ -949,6 +978,7 @@ validate_smtp_dsn(char const *dsn,
 *                   unix:/path/to/sock
 *                   local:/path/to/sock
 *                   inet:port  (host defaults to LOOPBACK)
+*                   inet_any:port (host defaults to INADDR_ANY)
 *  backlog -- listen backlog.  If -1, we use default of 5
 * %RETURNS:
 *  A listening UNIX-domain or TCP socket
@@ -964,11 +994,20 @@ make_listening_socket(char const *str, int backlog)
     struct sockaddr_un addr_un;
     int opt;
     int port;
+    uint32_t bind_addr;
 
     if (backlog <= 0) backlog = 5;
 
-    if (!strncmp(str, "inet:", 5)) {
-	path = str+5;
+    if (!strncmp(str, "inet:", 5) ||
+	!strncmp(str, "inet_any:", 9)) {
+	if (!strncmp(str, "inet:", 5)) {
+	    path = str+5;
+	    bind_addr = htonl(INADDR_LOOPBACK);
+	} else {
+	    path = str+9;
+	    bind_addr = htonl(INADDR_ANY);
+	}
+
 	if (sscanf(path, "%d", &port) != 1 ||
 	    port < 1 ||
 	    port > 65535) {
@@ -993,7 +1032,7 @@ make_listening_socket(char const *str, int backlog)
 
 	addr_in.sin_family = AF_INET;
 	addr_in.sin_port = htons(port);
-	addr_in.sin_addr.s_addr = (uint32_t) htonl(INADDR_LOOPBACK);
+	addr_in.sin_addr.s_addr = bind_addr;
 	if (bind(sock, (struct sockaddr *) &addr_in, sizeof(addr_in)) < 0) {
 	    syslog(LOG_ERR, "make_listening_socket: bind: %m");
 	    close(sock);
@@ -1107,3 +1146,27 @@ connect_to_socket(char const *str)
     }
     return sock;
 }
+/**********************************************************************
+* %FUNCTION: do_delay
+* %ARGUMENTS:
+*  sleepstr -- Number of seconds to delay as an ASCII string.
+* %RETURNS:
+*  Nothing
+* %DESCRIPTION:
+*  Sleeps for specified number of seconds
+***********************************************************************/
+void
+do_delay(char const *sleepstr)
+{
+    int snooze;
+    if (!sleepstr || !*sleepstr) {
+	return;
+    }
+
+    if (sscanf(sleepstr, "%d", &snooze) != 1 || snooze <= 0) {
+	return;
+    }
+    while(snooze) {
+	snooze = sleep(snooze);
+    }
+}
